#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
  Code:        abaqus_input.py

  Author:      Jerawan Armstrong
               Monte Carlo Codes (XCP-3) Group
               X Computational Physics (XCP) Division
               Los Alamos National Laboratory

  Reference: This Python3 code was extracted from LA-UR-20-27139,
             approved for public release.

  Copyright (c) 2020 Triad National Security, LLC.  All rights reserved.

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE. 
"""
import numpy as np
import os

element_types  = ['c3d4', 'c3d6', 'c3d8', 'c3d10', 'c3d15', 'c3d20', 'sc8']

elset_keywords = ['material', 'statistic', 'source', 'matprop']

#------------------------------------------------------------------------------
class AbaqusInp:
    """ Abaqus Input """
    def __init__(self):
        self.heading   = ''
        self.parts     = [] 
        self.instances = []
        self.materials = []

    def set_heading(self,head):
        self.heading = head

    def get_heading(self):
        return self.heading

    def set_parts(self,parts):
        self.parts = parts

    def get_parts(self):
        return self.parts

    def set_instances(self,instances):
        self.instances = instances

    def get_instances(self):
        return self.instances

    def set_materials(self,materials):
        self.materials = materials

    def get_materials(self):
        return self.materials


#------------------------------------------------------------------------------
def read_abaqus(filename):
    if not os.path.isfile(filename):
        raise OSError('file does not exist: {:s}'.format(filename))
    else:
        print("\nRead an Abaqus input file: {:s}\n".format(filename))
        with open(filename, "r", encoding='utf-8',errors='ignore') as f:
            names, inpdata, inplines  = _read_buffer(f)

        return names, inpdata, inplines

#------------------------------------------------------------------------------   
def _read_buffer(f):
    
    # ----- BEGIN READING HEADING -----
    heading      = ''
    need_newline = False
    while True:
        line = f.readline()
        if not line: raise EOFError('read heading')

        if line.startswith('**'): continue

        if line.startswith('*'):
            lline = line.lower()

            if lline.startswith('*part'): break

            if lline.startswith('*heading'):
                line = f.readline()
                if not line: raise EOFError('read heading data')
                
                lline = line.lower()
                if lline.startswith('*part'): break
        
                if not line.startswith('**'): heading = line.strip()
                need_newline = True
                break
    # ----- END READING HEADING -----


    # ----- BEGIN READING PARTS -----
    part_names = []
    part_lines = {}
    part_data  = {}
    read_flag  = 0

    while True:
        if need_newline:
            line  = f.readline()
            if not line: raise EOFError('read part')
        else:
            need_newline = True

        if line.startswith('**'): 
            continue
        elif line.startswith('*'):
            lline = line.lower()
        else:
            continue
        
        #---------------
        #      Part
        #---------------
        if read_flag == 0 and lline.startswith('*part'):
            pname = _extract_value(line[6:],'name') 

            if len(pname) > 0:
                if pname in part_names:
                    raise ValueError('part name is not unique: {:s}'.format(pname)) 
                part_names.append(pname)
                pdata     = []
                plines    = []
                read_flag = 1
            else:
                raise ValueError('{:s} does not have a part name'.format(line)) 

        #-----------------
        #      Node 
        #-----------------
        elif read_flag == 1 and lline.startswith('*node'):
            temp_lines = []

            # ----- [begin] read node data -----
            while True:
                line = f.readline()
                if not line: raise EOFError('read node data of part {:s}'.format(pname))
                
                if line.startswith('*'):
                    lline = line.lower()
                    break
                else:
                    temp_lines.append(line)
            # ----- [end] read node data -----

            n = len(temp_lines)
            if n > 0:
                # pdata[0]
                pdata.append(n)  # number of nodes
                # plines[0]
                plines.append(temp_lines)
                
                read_flag    = 2
                need_newline = False
            else:
                raise ValueError('no node data for part {:s}'.format(pname))
  
        #-------------------------
        #      Element Type
        #-------------------------
        elif read_flag == 2 and lline.startswith('*element'):
            n_part_elements = 0  # number of elements in this part
            element_data    = []
            temp_lines      = []

            # ----- [begin] check element types -----
            while True:     
                tvalue = _extract_value(lline[9:], 'type')

                if len(tvalue) == 0 :
                    raise ValueError('{:s} does not an element type'.format(line))
                else:
                    etype = _extract_etype(tvalue)

                    if len(etype) == 0:
                        raise ValueError('invalid element type: {:s}'.format(line))
                    else:
                        
                        cur_temp_lines = []
                        
                        # ----- [begin] read element data -----
                        while True:
                            line = f.readline()
                            if not line: raise EOFError('read element data for part {:s}'.format(pname))
                
                            if line.startswith('*'):
                                lline = line.lower()
                                break
                            else:
                                cur_temp_lines.append(line)
                        # ----- [end] read element data -----

                        n = len(cur_temp_lines)
                        if n > 0:
                            if etype == 'c3d20':
                                if n%2 == 0:
                                    n = int(n/2)
                                    element_data.append([etype, n])
                                else:
                                    raise ValueError('bad number of C3D8 elements in part: {:}'.format(pname))
                            else:
                                element_data.append([etype, n])
                                
                            n_part_elements += n
                            temp_lines.append(cur_temp_lines)
                            
                        else:
                            raise ValueError('no element data for element type {:s} in part {:s}'.format(etype,pname))
                
                        if not lline.startswith('*element,'): break

                        if len(element_data)==2:
                            raise ValueError('number of *element lines > 2 in part {;}'.format(pname))        
            # ----- [end] check element types -----

            if len(element_data) == 0:
                raise ValueError('no element data in part {:s}'.format(pname))

            if len(element_data) == 2:
                n = [0, 0, 0, 0, 0, 0, 0]
                for d in element_data:
                    etype = d[0]
                    if    etype == 'c3d4'  : n[0] = n[0]+1
                    elif  etype == 'c3d6'  : n[1] = n[1]+1
                    elif  etype == 'c3d8'  : n[2] = n[2]+1
                    elif  etype == 'c3d10' : n[3] = n[3]+1
                    elif  etype == 'c3d15' : n[4] = n[4]+1
                    elif  etype == 'c3d20' : n[5] = n[5]+1
                    elif  etype == 'sc8'   : n[6] = n[6]+1
                        
                for i in n:
                    if i==2:
                        raise ValueError('two lines of the same element type in a part is not allowed; part name {:s}'.format(pname))
                    
                if n[0]==1:
                    raise ValueError('c3d4 type cannot mix with another element type; part name {:s}'.format(pname))

                if n[1]==1:
                    if n[4]==1:
                        raise ValueError('c3d6 type cannot mix with c3d15 type; part name {:s}'.format(pname))
                    if n[5]==1:
                        raise ValueError('c3d6 type cannot mix with c3d20 type; part name {:s}'.format(pname))
                    if n[6]==1:
                        raise ValueError('c3d6 type cannot mix with sc8 type; part name {:s}'.format(pname))

                if n[2]==1:
                    if n[4]==1:
                        raise ValueError('c3d8 type cannot mixed with c3d15 type; part name {:s}'.format(pname))
                    if n[5]==1:
                        raise ValueError('c3d8 type cannot mixed with c3d20 type; part name {:s}'.format(pname))

                if n[3]==1:
                    raise ValueError('c3d10 type cannot mixed with another element type; part name {:s}'.format(pname))

                if n[4]==1:
                    if n[6]==1:
                        raise ValueError('c3d15 type cannot mixed with sc8 type; part name {:s}'.format(pname))

            # pdata[1]
            pdata.append(element_data)  # [ [element_type, number_of_elements], ....]
                
            # plines[1]
            plines.append(temp_lines)
                
            read_flag    = 3
            need_newline = False
       
            
        # ------------ Elset W/O Internal ---------------------
        elif read_flag == 3 and lline.startswith('*elset') and lline.find('internal') == -1:
            n_part_mat_elsets     = 0  # number of material elsets in this part
            n_part_matprop_elsets = 0  # number of material property elsets in this part
            n_part_stat_elsets    = 0  # number of stat elsets in this part
            n_part_src_elsets     = 0  # number of source elsets i this part

            keyword_lines   = []
            elset_data      = []
            temp_lines      = []
           
            # ----- [begin] check eleset keywords -----
            while True:
                keywords = _extract_keywords(lline[7:])
                if len(keywords) == 0:
                    elset_flag = False
                else:
                    elset_flag = True

                #---------- begin elset_flag --------------------
                if elset_flag:
                    elset_keyword_line = line
                    elset_name         = _extract_value(line[7:], 'elset')
                    elset_num          = _extract_number(elset_name)
                    if elset_num is None:
                        raise ValueError('elset name does not ending with a number: {:s}'.format(elset_name))
                    elif elset_num <= 0:
                        raise ValueError('elset number <= 0 for elset name : {:s}'.format(elset_name))
                    
                    generate_index = lline.rfind('generate')
                   
                    # ----- [begin] read elset data -----
                    cur_temp_lines = []
                    while True:
                        line = f.readline()
                        if not line:  raise EOFError('read elset data of part {:s}'.format(pname))

                        if line.startswith('_'): break
                        if line.startswith('*'): break
                            
                        cur_temp_lines.append(line)

                        if generate_index > 0:
                            this_elset_line   = np.int_(line.strip().split(','))
                            n_this_elsets     = int((this_elset_line[1]-this_elset_line[0]+1)/this_elset_line[2])
                        else:
                            this_elset_line   = line.strip().split(',')
                            n_this_elsets     = len(this_elset_line)

                        if 'material'  in keywords: n_part_mat_elsets      += n_this_elsets
                        if 'matprop'   in keywords: n_part_matprop_elsets  += n_this_elsets
                        if 'statistic' in keywords: n_part_stat_elsets     += n_this_elsets
                        if 'source'    in keywords: n_part_src_elsets      += n_this_elsets
                        # ----- [end] read elset data -----
                        
                    n = len(cur_temp_lines)
                    if n > 0:
                        keyword_lines.append(elset_keyword_line.strip()) 
                        elset_data.append([keywords, elset_num, generate_index, n])
                        temp_lines.append(cur_temp_lines)
            
                        if line.startswith('*'):
                            lline = line.lower()
                            if lline.startswith('*elset,') and lline.find('internal') == -1: continue
                            if lline.startswith('*end part'): break
                    else:
                        raise ValueError('no elset data for elset line {:s} of part {:s}'.format(elset_keyword_line,pname))      
                #---------- end elset_flag --------------------
        
                endpart_flag = False
                while True:
                    line = f.readline()
                    if not line:  raise EOFError('read elset data in a part')
                    
                    if line.startswith('*'):
                        lline = line.lower()
                        if lline.startswith('*elset,') and lline.find('internal') == -1:
                            endpart_flag = False
                            break
                        if lline.startswith('*end part'):
                            endpart_flag = True
                            break

                if endpart_flag: break
            # ----- [end] check eleset keywords -----

            #--------------------------------
            # check elset data in this part
            # -------------------------------
            if len(elset_data) == 0:
                raise ValueError('no elset data for part {:s}'.format(pname)) 

            elset_mat_nums       = []
            elset_matprop_nums   = []
            elset_statistic_nums = []
            elset_source_nums    = []
            
            mat_elsets     = 0  # number of material elset keyword lines in this part
            matprop_elsets = 0  # number of matprop elset keyword lines in this part
            stat_elsets    = 0  # number of statistic elset keyword lines in this part
            source_elsets  = 0  # number of source elset keyword lines in this part
            
            for elset in elset_data:
                keywords  = elset[0]
                elset_num = elset[1]
                
                if 'material' in keywords:
                   elset_mat_nums.append(elset_num)
                   mat_elsets     += 1
                        
                if 'matprop' in keywords:
                    if elset_num in elset_matprop_nums:
                        raise ValueError('part {:s} has multiplie elset matprop number {:d}'.format(pname, elset_num))
                    else:
                        elset_matprop_nums.append(elset_num)
                        matprop_elsets += 1
                        
                if 'statistic' in keywords:
                    elset_statistic_nums.append(elset_num)
                    stat_elsets += 1
                        
                if 'source' in keywords: 
                    if elset_num in elset_source_nums:
                        raise ValueError('part {:s} has multiple elset source number {:d}'.format(elset_num,pname))
                    else:
                        elset_source_nums.append(elset_num)
                        source_elsets += 1
                
           
            if mat_elsets == 0:
                raise ValueError('no material elset in part {:s}'.format(pname))

            # all elements in each part must be assigned to a material
            if n_part_mat_elsets < n_part_elements:
                raise ValueError('not all elements in part {:s} are assigned to a material'.format(pname))
            elif n_part_mat_elsets > n_part_elements:
                raise ValueError('number of elements for material elsets > number of elements in part {:s}'.format(pname))

                          
            if matprop_elsets > 0:
                if mat_elsets > 1:
                    for elset in elset_data:
                        keywords = elset[0]
                        if 'matprop' in keywords and 'material' in keywords: continue
                        raise ValueError('multiple material properties and multiple materials in a part is not allowed')
   
                if stat_elsets > 1:
                    for elset in elset_data:
                        keywords = elset[0]
                        if 'matprop' in keywords and 'material' in keywords and 'statistic' in keywords: continue
                        raise ValueError('multiple material properties and multiple statistics in a part is not allowed')

                if source_elsets > matprop_elsets:
                    raise ValueError('number of source elsets > number of matprop elsets in part {:s}'.format(pname))

                if n_part_matprop_elsets < n_part_elements:
                    print("number of matprop elsets = {:d}".format(n_part_matprop_elsets))
                    print("number of elements in this part = {:d}".format(n_part_elements))
                    raise ValueError('not all elements in part {:s} are assigned to a material property'.format(pname))
                elif n_part_matprop_elsets > n_part_elements:
                    print("number of matprop elsets = {:d}".format(n_part_matprop_elsets))
                    print("number of elements in this part = {:d}".format(n_part_elements))
                    raise ValueError('number ot elements for material property elsets > number of elements in part {:s}'.format(pname))
                
            else:
                if stat_elsets == 0:
                    raise ValueError('no statisitc elset in part {:s}'.format(pname))

                if mat_elsets > stat_elsets:
                    raise ValueError('not all materials in part {:s} are assigned to statistics regions'.format(pname))

                if source_elsets > stat_elsets:
                    raise ValueError('number of source elsets > number of statistics elsets in part {:s}'.format(pname))
                
                if n_part_stat_elsets < n_part_elements:
                    raise ValueError('not all elements in part {:s} are assigned to a statistics'.format(pname))
                elif n_part_matprop_elsets > n_part_elements:
                    raise ValueError('number ot elements for statistic elsets > number of elements in part {:s}'.format(pname))

            # pdata[2]
            pdata.append(elset_data)
            
            # plines[2]
            plines.append(temp_lines)
            
            # pdata[3]
            pdata.append([mat_elsets, matprop_elsets, stat_elsets, source_elsets])

            # pdata[4]
            pdata.append([n_part_mat_elsets, n_part_matprop_elsets, n_part_stat_elsets, n_part_src_elsets])

            # pdata[5]
            pdata.append([elset_mat_nums, elset_matprop_nums, elset_statistic_nums, elset_source_nums])

            # pdata[6]
            pdata.append(keyword_lines)
            
            #------------------------------------------------------------------------------------------------------------------
            # pdata is a list
            #    pdata[0] = number_of_nodes
            #
            #    pdata[1] = element_data, element_data is a list with length n where n is a number of element types in a part
            #       
            #               element_data[i] = [etype, number_of_elements]
            #        
            #    pdata[2] = elset_data, elset_data is a list with length n where n is a number of elsets in a part
            #
            #               elset_data[i] = [keywords,elset_number, generate_index, number_of_elset_lines]
            #
            #               keywords is a list with a length 1, 2, 3, or 4 depending on the elset name. For example,
            #                               
            #                  keywords = ['material']  
            #                             for *Elset, elset=????material????-xxxx
            #                               
            #                  keywords = ['material', 'statistic']  
            #                             for *Elset, elset=????material-statistic????-xxxx
            #
            #                  keywords = ['material', 'statistic', 'source'] 
            #                             for *Elset, elset=????material-statistic-source????-xxxx
            #           
            #                  keywords = ['material', 'staitctic, 'soource', 'matprop'] 
            #                             for *Elset, elset=????material-statistic-source-source????-xxxx
            #
            #    pdata[3] = [mat_elsets, matprop_elsets, stat_elsets, source_elsets]
            #
            #    pdata[4] = [n_part_mat_elsets, n_part_matprop_elsets, n_part_stat_elsets, n_part_src_elsets]
            #   
            #    pdata[5] = [elset_mat_nums, elset_matprop_nums, elset_statistic_nums, elset_source_nums]
            # 
            #    pdata[6] = keyword_lines = [elset_keyword_line1, elset_keyword_line2, ...]
            #
            #  plines is a list of data lines in a part. There are 3 slices of data
            #         plines[0] = node data
            #         plines[1] = element data
            #         plines[2] = elset data 
            #
            # ---------------------------------------------------------------------------------------------------------------
            part_data[pname]  = pdata
            part_lines[pname] = plines
            read_flag         = 0
        
        
        # ------------ END PART -------------
        elif lline.startswith('*end part'):
            if read_flag==0:
                raise ValueError('mismatch *part and *end part')
            elif read_flag==1:
                raise ValueError('no node data in part {:s}'.format(pname))
            elif read_flag==2:
                raise ValueError('no element data in part {:s}'.format(pname))
            elif read_flag==3:
                raise ValueError('no elset data in part {:s}'.format(pname))
                
    
        # --------- ASSEMBLY ------------------
        elif lline.startswith('*assembly'):
            if read_flag == 0:
                need_newline = False
                break
            else:
                raise ValueError('bad Abaqus input format when reading *assembly')
           
        # ------------ MATERIAL ----------------
        elif lline.startswith('*material,'):
            if read_flag == 0:
                need_newline = False
                break
            else:
                raise ValueError('bad Abaqus input format when reading *material')
    #---------------- END READING PARTS ---------------------

    temp_line = []

    if len(part_names)==0: raise ValueError('no part data in this file')
    if read_flag != 0: raise ValueError('mismatch *part and *end part')


    #----------- BEGIN READING ASSEMBLY & MATERIALS ----------
    num_assembly     = 0
   
    instance_names   = []
    instance_lines   = {}
    instance_data    = {}

    material_names   = []
    material_nums    = []
    material_data    = {}

    assembly_flag    = 0
    material_flag    = 0

    while True:
        if need_newline:
            line  = f.readline()
            if not line: break
        else:
            need_newline = True

        if line.startswith('**'): 
            continue
        elif line.startswith('*'):
            lline = line.lower()
        else:
            continue
        
        # --------- *ASSEMBLY -----------
        if lline.startswith('*assembly'):
            if num_assembly == 0:
                num_assembly  = num_assembly + 1
                need_newline  = True
                assembly_flag = 1
                material_flag = 0
            else:
                raise ValueError('file has more than 1 assembly')
         
        # ---------- * INSTANCE -----------
        elif assembly_flag == 1 and lline.startswith('*instance,'):

            ins_name = _extract_value(line[10:],'name')
            if len(ins_name) == 0 :
                raise ValueError('no instance name {:s}'.format(line))
            else:
                if ins_name in instance_names:
                    raise ValueError('instance name is not unique {:s}'.format(line))
                else:
                    pname = _extract_value(line[10:],'part')
                    if pname not in part_names:
                        raise ValueError("instance's partname does not match with any partname. \n{:s}".format(line))

                    else:
                        inslines = []
                        ndata    = 0
                        while True:
                            line  = f.readline()
                            if not line:  raise EOFError('read instance data {:s}'.format(ins_name))
                
                            if line.startswith('*'):
                                if line.lower().startswith('*end instance'): 
                                    break
                                else:
                                    raise ValueError('bad instance data; see {:s}'.format(line))

                            else:
                                if ndata >= 2: 
                                    raise ValueError('bad instance data; see {:s}'.format(line))
                                else:
                                    inslines.append(line)
                                    ndata = ndata + 1
                                
                        instance_names.append(ins_name)
                        instance_data[ins_name]  = [pname, ndata]   
                        instance_lines[ins_name] = inslines
                        need_newline             = True
                        
        # -------- *END ASSEMBLY --------------
        elif lline.startswith('*end assembly'):
            assembly_flag = 0
            material_flag = 0
            need_newline  = True

        # ----------- *MATERIAL ----------------
        elif lline.startswith('*material,'):
            mat_name = _extract_value(line[10:],'name')
            if len(mat_name) == 0 :
                raise ValueError('no material name {:s}'.format(line))
            else:
                if mat_name in material_names:
                    raise ValueError('material name is not unique {:s}'.format(mat_name))
                else:                   
                    mat_num = _extract_number(mat_name)
                    if mat_num is None:
                        raise ValueError('material name is not ending with number')
                    else:
                        if mat_num is material_nums:
                           raise ValueError('material number is not unique {:s}'.format(mat_name))
                        material_names.append(mat_name)
                        material_nums.append(mat_num)
                        material_data[mat_num] = [mat_name, 0.0]
                        material_flag           = 1
                        need_newline            = True

        elif lline.startswith('*density'):
            if material_flag == 0:
                raise ValueError('mismatch material and its density')
            else:
                line  = f.readline()
                if not line:  raise EOFError('read density data')
                
                densities = line.split(',')
               
                if len(densities) == 0:
                    raise ValueError('bad density value: {:s}'.format(line))
                else:
                    density = densities[0].strip()
                    try:
                        density                 = float(density)
                        material_data[mat_num] = [mat_name, density]
                        material_flag           = 0
                    except:
                        raise ValueError('density value is not a number: {:f}'.format(line))
                   
    #----------- END READING ASSEMBLY & MATERIALS ----------
    if len(instance_names) == 0: 
        raise  ValueError('no instance in this file')
    if len(material_names) == 0:
        raise ValueError('no material in this file')
     
    inpdata  = AbaqusInp()
    inplines = AbaqusInp()

    inpdata.set_heading(heading)
    inpdata.set_parts(part_data)
    inpdata.set_instances(instance_data)
    inpdata.set_materials(material_data)

    inplines.set_heading('')
    inplines.set_parts(part_lines)
    inplines.set_instances(instance_lines)
    inplines.set_materials('')

    names = [part_names, instance_names, material_names]  

    return [names, inpdata, inplines]
   
#-----------------------------------------------------------------------------       
def _extract_value(line,option):
    """ 
    Extract value of option from line.

    Example: line   = '*Part, name=" part-name "'
             option = Name
             value  = 'part-name' 
    """
    value = ''

    line_list = line.split(',')

    for l in line_list:
        i = l.lower().find(option.lower())
        if i >= 0:
            j = l[i:].find('=')
            if j > 0: 
                value = l[i+j+1:].strip().strip('"')
                break

    return value

#------------------------------------------------------------------------------
def _extract_etype(value):
    """
    Extract etype from value

    Example: value = 'C3D8R'
             etype = 'c3d8'
    """
    etype = ''
    for et in element_types:
        i = value.lower().find(et)
        if i >= 0 :
            etype = et
            break

    return etype

#------------------------------------------------------------------------------
def _extract_keywords(value):
    """
    Extract keywords from value

    Example: value = 'set-material-statistic-001'
             keywords = [material, stattistic]
    """
    keywords = []
    for kw in elset_keywords:
        i = value.lower().find(kw)
        if i >= 0: keywords.append(kw)

    if 'statistic' not in keywords:
        i = value.lower().find('tally')
        if i >= 0: keywords.append('statistic')

    return keywords

#------------------------------------------------------------------------------
def _extract_number(value):
    """
    Extract number from value.
    
    Example: value ='Set-material-001'
             number   = 1
    """
    number = None
    i = value.rfind('-')
    j = value.rfind('_')
    k = max(i,j)
    if k >= 0:
        if value[k+1:].isnumeric():
            number = int(value[k+1:])
        else:
            raise ValueError('{:s} does not end with a number'.format(value))
    
    return number

#------------------------------------------------------------------------------
def extract_node_data(line):
    line = line.strip().split(',')
    return int(line[0]), list(np.float_(line[1:]))

#------------------------------------------------------------------------------
def extract_element_data(line):
    line = line.strip().split(',')
    if '' in line: line.remove('')
    return list(np.int_(line))

#------------------------------------------------------------------------------
def extract_elset_data_generate(line):
    line = line.strip().split(',')
    return list(np.int_(line))

#------------------------------------------------------------------------------
def extract_elset_data(line):
    line = line.strip().split(',')
    if '' in line: line.remove('')
    return list(np.int_(line))
            
#------------------------------------------------------------------------------
def extract_instance_data(line):
    line = line.strip().split(',')
    return list(np.float_(line))

#------------------------------------------------------------------------------
def extract_mesh_data(names, inpdata, inplines, dentype):
    """
    Extract mesh data
    """
    part_names      = names[0]
    instance_names  = names[1]
    material_names  = names[2]
    
    heading       = inpdata.get_heading()
    part_data     = inpdata.get_parts()
    instance_data = inpdata.get_instances()
    material_data = inpdata.get_materials()
    
    part_lines     = inplines.get_parts()
    instance_lines = inplines.get_instances()
    
    
    mesh_info = ''
    if len(heading) > 0:
        mesh_info = mesh_info + '\n   heading: {:s}\n'.format(heading)

    mesh_info = mesh_info + '\n   number of parts: {:d}'.format(len(part_names))
    mesh_info = mesh_info + '\n   number of instances: {:d}'.format(len(instance_names))
    mesh_info = mesh_info + '\n   number of materials: {:d}\n'.format(len(material_names))

    total_nodes    = 0
    total_elements = []
    for i in range(len(element_types)): total_elements.append(0)

    for pname in part_names:
        pdata = part_data.get(pname)
        total_nodes += pdata[0]
        
        element_data = pdata[1]
        for el in element_data:
            etype = el[0]
            nels  = el[1]
            if   etype == 'c3d4':  total_elements[0] += nels
            elif etype == 'c3d6':  total_elements[1] += nels
            elif etype == 'c3d8':  total_elements[2] += nels
            elif etype == 'c3d10': total_elements[3] += nels
            elif etype == 'c3d15': total_elements[4] += nels
            elif etype == 'c3d20': total_elements[5] += nels
            elif etype == 'sc8':   total_elements[6] += nels
            else:
                raise ValueError('invalid element type :{:s}'.format(etype))
            
    sum_elements = np.sum(total_elements)
    mesh_info = mesh_info + '\n   total  number of nodes: {:d}'.format(total_nodes)
    mesh_info = mesh_info + '\n   total  number of elements: {:d}\n'.format(sum_elements)
    if total_elements[0] > 0:
        mesh_info = mesh_info + '\n   total  number of linear tet elements: {:d}'.format(total_elements[0])
    if total_elements[1] > 0:
        mesh_info = mesh_info + '\n   total  number of linear pent elements: {:d}'.format(total_elements[1])
    if total_elements[2] > 0:
        mesh_info = mesh_info + '\n   total  number of linear hex elements: {:d}'.format(total_elements[2])
    if total_elements[3] > 0:
        mesh_info = mesh_info + '\n   total  number of quad tet elements: {:d}'.format(total_elements[3])
    if total_elements[4] > 0:
        mesh_info = mesh_info + '\n   total  number of quad pent elements: {:d}'.format(total_elements[4])
    if total_elements[5] > 0:
        mesh_info = mesh_info + '\n   total  number of quad hex elements: {:d}'.format(total_elements[5])
    if total_elements[6] > 0:
        mesh_info = mesh_info + '\n   total  number of sc8 elements: {:d}'.format(total_elements[6])
    mesh_info = mesh_info + '\n'
        
    for pname in part_names:
        pdata    = part_data.get(pname)
        plines   = part_lines.get(pname)
        
        n_nodes   = pdata[0]
        mesh_info = mesh_info + '\n\n   *** part name: {:s}'.format(pname)
        mesh_info = mesh_info + '\n   *** number of nodes: {:d}'.format(n_nodes)
 
        element_data = pdata[1]
        for el in element_data:
            etype = el[0]
            nels  = el[1]
            mesh_info = mesh_info + '\n   *** element_type: {:s}'.format(etype)
            mesh_info = mesh_info + '\n       number of elements: {:d}'.format(nels)

        sp = '-'
        elset_data  = pdata[2]
        elset_lines = plines[2]
        
        for elset, lines  in zip(elset_data,elset_lines):
            keywords              = elset[0]
            elset_num             = elset[1]
            generate_index        = elset[2]
            
            if generate_index >= 1:
                line      = lines[0]
                line_data = extract_elset_data_generate(line)
                nelsets   = int((line_data[1] - line_data[0] + 1)/ line_data[2])
                
            else:
                nelsets = 0
                for line in lines:
                    line_data = extract_elset_data(line)
                    nelsets  += len(line_data)
            
            elset_keywords = sp.join(keywords)
            mesh_info = mesh_info + '\n   *** elset keywords: {:s}'.format(elset_keywords)
            mesh_info = mesh_info + '\n       elset number: {:d}'.format(elset_num)
            mesh_info = mesh_info + '\n       number of elements in this set: {:d}'.format(nelsets)
            
        mesh_info = mesh_info + '\n'
        
    for ins_name in instance_names:
        ins_data  = instance_data.get(ins_name)
        ins_lines = instance_lines.get(ins_name)
        
        mesh_info = mesh_info + '\n\n   *** instance name: {:s}'.format(ins_name)
        mesh_info = mesh_info + '\n       part name: {:s}'.format(ins_data[0])
           
        if ins_data[1]==0:
            mesh_info = mesh_info + '\n       translation: no'
            mesh_info = mesh_info + '\n       rotation: no'

        elif ins_data[1]==1:
            mesh_info = mesh_info + '\n       translation: yes'
            mesh_info = mesh_info + '\n       rotation: no'
            mesh_info = mesh_info + '\n       {:s}'.format(ins_lines[0])

        elif ins_data[1]==2:
            mesh_info = mesh_info + '\n       translation: yes'
            mesh_info = mesh_info + '\n       rotation: yes'
            mesh_info = mesh_info + '\n       {:s}'.format(ins_lines[0])
            mesh_info = mesh_info + '\n       {:s}'.format(ins_lines[1])
    mesh_info = mesh_info + '\n'
                 
    material_nums = material_data.keys()

    for m in material_names:
        for n in material_nums:
            mdata = material_data.get(n)
            if mdata[0]==m: break
        mesh_info = mesh_info + '\n\n   *** material name: {:s}'.format(m)
        mesh_info = mesh_info + '\n       material number: {:d}'.format(n)
        if dentype == 'mass_den':
            mesh_info = mesh_info + '\n       material density: {:10.5f}'.format(mdata[1])
        elif dentype == 'atom_den':
            mesh_info = mesh_info + '\n       material density: {:10.5e}'.format(mdata[1])
            
    mesh_info = mesh_info + '\n'
    
    return mesh_info

#------------------------------------------------------------------------------
def compute_translated_points(begin_points,end_points):
    begin_points = np.array(begin_points)
    end_points   = np.array(end_points)
    return  begin_points + end_points

#------------------------------------------------------------------------------
def compute_rotated_points(points, begin_points, end_points, angle):
    points       = np.array(points)
    begin_points = np.array(begin_points)
    end_points   = np.array(end_points)

    angle    = angle*np.pi/180.0 # convert degree to radian
    cos_ang  = np.cos(angle)
    sin_ang  = np.sin(angle)
    a        = end_points - begin_points

    a2                = a*a
    a_time_sin_ang    = a*sin_ang
    a01               = a[0]*a[1]
    a02               = a[0]*a[2]
    a12               = a[1]*a[2]
    one_minus_cos_ang = 1.0 - cos_ang

    m = np.matrix('0 0 0; 0 0 0; 0 0 0')

    m[0,0] = cos_ang + one_minus_cos_ang * a2[0]
    m[0,1] = one_minus_cos_ang * a01 - a_time_sin_ang[2]
    m[0,2] = one_minus_cos_ang * a02 + a_time_sin_ang[1]

    m[1,0] = one_minus_cos_ang * a02  + a_time_sin_ang[2]
    m[1,1] = cos_ang + one_minus_cos_ang * a2[1]
    m[1,2] = one_minus_cos_ang * a12  - a_time_sin_ang[0]

    m[2,0] = one_minus_cos_ang * a02 - a_time_sin_ang[1]
    m[2,1] = one_minus_cos_ang * a12 + a_time_sin_ang[0]
    m[2,2] = cos_ang + one_minus_cos_ang * a2[2]

    x = points - begin_points
    x = x.reshape((3,1))
    
    return np.array(m*x).reshape((3,))+ begin_points

#------------------------------------------------------------------------------
def compute_global_model_extents(part_names, part_data, part_lines,
                                 instance_names, instance_data, instance_lines):
    """
    Compute global model extents
    """
    # process part data
    part_xpoints = {}
    part_ypoints = {}
    part_zpoints = {}

    for pname in part_names:
        plines = part_lines.get(pname)

        xpoints = []
        ypoints = []
        zpoints = []

        node_lines = plines[0]
        for line in node_lines:
            id, xyz = extract_node_data(line)
            xpoints.append(xyz[0])
            ypoints.append(xyz[1])
            zpoints.append(xyz[2])
        
        part_xpoints[pname] = [np.min(xpoints), np.max(xpoints)]
        part_ypoints[pname] = [np.min(ypoints), np.max(ypoints)]
        part_zpoints[pname] = [np.min(zpoints), np.max(zpoints)]

    
    # process instance data
    xmin_points = []
    xmax_points = []

    ymin_points = []
    ymax_points = []

    zmin_points = []
    zmax_points = []

    for ins_name in instance_names:
        ins_data  = instance_data.get(ins_name)
        ins_lines = instance_lines.get(ins_name)

        pname  = ins_data[0]
        nlines = ins_data[1]
           
        xmin, xmax = part_xpoints[pname]
        ymin, ymax = part_ypoints[pname]
        zmin, zmax = part_zpoints[pname]

        if nlines==1: # translation
            t  = extract_instance_data(ins_lines[0])
            if np.any(t != 0.0):
                pmin = [xmin,ymin,zmin]
                pmax = [xmax,ymax,zmax]
                xmin, ymin, zmin  = compute_translated_points(pmin,t)
                xmax, ymax, zmax  = compute_translated_points(pmax,t) 
                            
        elif nlines==2: # translation and rotation
            t  = extract_instance_data(ins_lines[0])
            if np.any(t != 0.0):
                pmin = [xmin,ymin,zmin]
                pmax = [xmax,ymax,zmax]
                xmin, ymin, zmin  = compute_translated_points(pmin,t)
                xmax, ymax, zmax  = compute_translated_points(pmax,t) 
                    
            r = extract_instance_data(ins_lines[1])
            if np.any(r !=0.0):
                p1    = r[0:3]
                p2    = r[3:6]
                angle = r[-1]

                pmin = [xmin,ymin,zmin]
                pmax = [xmax,ymax,zmax]
                xmin, ymin, zmin = compute_rotated_points(pmin, p1, p2, angle)
                xmax, ymax, zmax = compute_rotated_points(pmax, p1, p2, angle)

        xmin_points.append(xmin)
        xmax_points.append(xmax)

        ymin_points.append(ymin)
        ymax_points.append(ymax)

        zmin_points.append(zmin)
        zmax_points.append(zmax) 
       
    xmax = np.max(xmax_points)
    xmin = np.min(xmin_points)
    
    ymax = np.max(ymax_points)
    ymin = np.min(ymin_points)
    
    zmax = np.max(zmax_points)
    zmin = np.min(zmin_points)

    return [xmin,xmax,ymin,ymax,zmin,zmax]
    
#------------------------------------------------------------------------------
