#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
  Code:        abq_part_data_checking.py   

  Author:      Jerawan Armstrong
               Monte Carlo Codes (XCP-3) Group
               X Computational Physics (XCP) Division
               Los Alamos National Laboratory

  Reference: This Python3 code was extracted from LA-UR-20-27139,
             approved for public release.

  Copyright (c) 2020 Triad National Security, LLC.  All rights reserved.

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE. 
"""
import numpy as np

#------------------------------------------------------------------------------
def check_node_element_data(part_names, part_node_data, part_element_data):
    """
    Check whether node ids in element connectivity in each part are matched
    with node ids from node data
    """
    for pname in part_names:
        node_data    = part_node_data.get(pname)
        element_data = part_element_data.get(pname)

        if node_data[1]:
            node_ids = node_data[2]
        else:
            node_ids = np.arange(1, node_data[0]+1, 1)
        node_ids = set(node_ids)

        n_els    = element_data[0]
        cons     = element_data[3]

        element_connects = set()
        
        for i in range(len(n_els)):
            if n_els[i] == 0: continue
            element_connects = element_connects.union(cons[i])
    
        if node_ids.intersection(element_connects) != element_connects:
            print("Part Name: {:s}".format(pname))
            print("List of bad node ids in element connectivity")
            #for n in element_connects:
            #    if n not in node_ids: print(n)
            raise ValueError('bad element connectivities in part {:s}'.format(pname))
        
#-----------------------------------------------------------------------------------
def check_material_elset_data(part_names,part_data,material_data):
    material_nums = material_data.keys()
    for pname in part_names:
        pdata = part_data.get(pname)
        elset_mat_nums = pdata[5][0]
        for nm in elset_mat_nums:
            if nm not in material_nums:
                raise ValueError('material number {:d} in part {:s} is not in a material list'.format(nm,pname))

#-----------------------------------------------------------------------------------              
def check_element_elset_data(part_element_ids, part_elset_element_data):

    names = part_elset_element_data.keys()
    
    for name in names:
        elset_element_data = part_elset_element_data.get(name)
        n_elements         = elset_element_data[0]
        if n_elements == 0: continue
        
        element_ids        = set(elset_element_data[1])
        pname              = elset_element_data[2]
        
        temp_element_ids   = part_element_ids.get(pname)

        if temp_element_ids.intersection(element_ids) != element_ids:
            print('list of bad element ids:')
            for e in element_ids:
                if e not in temp_element_ids: print(e)
            raise ValueError('bad element elset in part {:s}'.format(pname))

            
        
        
        
