#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
  Code:        write_mcnp_um_input.py

  Author:      Jerawan Armstrong
               Monte Carlo Codes (XCP-3) Group
               X Computational Physics (XCP) Division
               Los Alamos National Laboratory

  Reference: This Python3 code was extracted from LA-UR-20-27139,
             approved for public release.

  Copyright (c) 2020 Triad National Security, LLC.  All rights reserved.

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE. 
"""
import numpy as np
import os

from abaqus_input import read_abaqus
from abaqus_input import compute_global_model_extents
from abaqus_input import extract_mesh_data

from abq_part_elsets import  compute_part_elset_mat_nums
from abq_part_elsets import  compute_part_elset_data
from abq_part_elsets import  compute_part_elset_material_data

from abq_part_data_checking import check_material_elset_data

#-------------------------------------------------------------------------------
def write_mcnp_um_input(filein, fileout=None, eeout=None, meshinfo=None,
                        cellcards=None, surfacecards=None, datacards=None,
                        bgmaterial=0, lenconv=1.0, radiusext=1.0, denunit='g/cc',
                        writemeshinfo=False, writecomments=False):

    print("\n***** Write MCNP skeleton input file *****")

    invalid_input = 0

    if not os.path.isfile(filein):
        print("\nfile does not exist: {:>s}\n".format(filein))
        invalid_input = 1

    if filein.lower() != filein:
        print('\nAn Abaqus input filename does not in lowercase: {:s}'.format(filein))
        print("MCNP converts filenames on EMBED cards into lowercase letters.")
        print("Abaqus input filename should be in lowercase letters to prevent")
        print("an error when running MCNP.\n")
        invalid_input = 2

    include_cell_cards = False
    if cellcards is not None:
        if not os.path.isfile(cellcards):
            print('\nThe cell card file does not exist: {:s}'.format(cellcards))
            invalid_input = 4
        else:
            include_cell_cards = True

    include_surface_cards = False
    if surfacecards is not None:
        if not os.path.isfile(surfacecards):
            print('\nThe surface card file does not exist: {:s}'.format(surfacecards))
            invalid_input = 5
        else:
            include_surface_cards = True
            
    include_data_cards = False
    if datacards is not None:
        if not os.path.isfile(datacards):
            print('\nThe data card file does not exist: {:s}'.format(datacards))
            invalid_input = 6
        else:
            include_data_cards = True

    if int(bgmaterial) < 0:
        print(">>>Error: invalid background material number; it must be non-negative")
        print("          input value = {:d}".format(bgmaterial))
        invalid_input = -1

    if lenconv <= 0.0:
        print(">>>Error: invalid length conversion; it must be positive")
        print("          input value = {:f}".format(lenconv))
        invalid_input =  -2

    if radiusext <= 0.0:
        print(">>>Error: invalid radius extension; it must be positive")
        print("          input value = {:f}".format(radiusext))
        invalid_input =  -3

    if denunit.lower() not in ['g/cc', 'g/cm^3', 'g/cm**3', 'atoms/barn-cm']:
        print(">>>Error: invalid density unit; denunit must be g/cm^3 or atoms/barn-cm")
        print("          input value = {:f}".format(radiusext))
        invalid_input =  -3
    else:
        if denunit.lower() in ['atoms/barn-cm','atoms/b-cm']: 
            dentype = 'atom_den'
        else:
            dentype = 'mass_den'
        
    if invalid_input != 0: return None


    name = os.path.basename(filein)
    name = name.split('.')
    
    if fileout is None: fileout = name[0]+'.mcnp'
    if os.path.isfile(fileout):
        print('\nThe MCNP input file exist: {:s}'.format(fileout))
        print('Manaully remove this file or give the new MCNP input file\n')
        invalid_input = 100

    if eeout is None:
        eeout = name[0]+'.eeout'
    else:
        eeout_lower = eeout.lower()
        if eeout != eeout_lower:
            print("\nUser's EEOUT filename: {:s}".format(eeout))
            print("EEOUT filename is changed to: {:s}".format(eeout_lower))
            eeout = eeout_lower

    if meshinfo is None:
        meshinfo = name[0]+'.info'
    if os.path.isfile(meshinfo):
        print('\nThe mesh information file exist: {:s}'.format(meshinfo))
        print('Manaully remove this file or give the new mesh information file\n')
        invalid_input = 101

    if invalid_input != 0: return None

    
    #-------------- Read Abaqus Input File ---------------
    names, inpdata, inplines = read_abaqus(filein)

    heading       = inpdata.get_heading()
    part_data     = inpdata.get_parts()
    instance_data = inpdata.get_instances()
    material_data = inpdata.get_materials()

    part_lines     = inplines.get_parts()
    instance_lines = inplines.get_instances()

    part_names      = names[0]
    instance_names  = names[1]
    material_names  = names[2]

    print("   Number of parts:     {:d}".format(len(part_names)))
    print("   Number of instances: {:d}".format(len(instance_names)))
    print("   Number of materials: {:d}\n".format(len(material_names)))

    if writemeshinfo:
        mesh_info = extract_mesh_data(names, inpdata, inplines, dentype)
    del names, inpdata, inplines

    
    #------- Process Abaqus Input File -------------
    print("Processing Abaqus input file")
    extents = compute_global_model_extents(part_names, part_data, part_lines,
                                           instance_names, instance_data, instance_lines)
    extents = np.array(extents)*lenconv
    
    xmin = extents[0]
    xmax = extents[1]
    ymin = extents[2]
    ymax = extents[3]
    zmin = extents[4]
    zmax = extents[5]

    # In um_pre_op fortran version, radius is hard coded as 0.6*diagonal. 
    x = xmax - xmin
    y = ymax - ymin
    z = zmax - zmin
    diagonal = np.sqrt(x*x+y*y+z*z)
    center   = 0.5*np.array([xmin+xmax, ymin+ymax, zmin+zmax])
    radius   = 0.5*diagonal + radiusext
    
    print("\n   Global Model Extents (in cm)")
    print("      Min X: {:20.5E}; Max X: {:20.5E}".format(xmin, xmax)) 
    print("      Min Y: {:20.5E}; Max Y: {:20.5E}".format(ymin, ymax))
    print("      Min Z: {:20.5E}; Max Z: {:20.5E}\n".format(zmin, zmax))

    text_info = "Abaqus Input File: {}".format(filein)

    text_info = text_info + "\n\nHigh-Level Mesh Information:"
    text_info = text_info + "\n   Number of Parts:     {:d}".format(len(part_names))
    text_info = text_info + "\n   Number of Instances: {:d}".format(len(instance_names))
    text_info = text_info + "\n   Number of Materials: {:d}".format(len(material_names))

    text_info = text_info + "\n\nGlobal Model Extents (in cm)"
    text_info = text_info + "\n   Min X: {:20.5E}; Max X: {:20.5E}".format(xmin, xmax)
    text_info = text_info + "\n   Min Y: {:20.5E}; Max Y: {:20.5E}".format(ymin, ymax)
    text_info = text_info + "\n   Min Z: {:20.5E}; Max Z: {:20.5E}\n".format(zmin, zmax)


    print("Creating MCNP pseudo-cells")

    part_elset_mat_nums = compute_part_elset_mat_nums(part_names, part_data, material_data) 
    part_elset_data     = compute_part_elset_data(part_names, part_data)
    check_material_elset_data(part_names,part_data,material_data)

    # set material elset data
    part_elset_material_data = compute_part_elset_material_data(part_names,
                                                                part_lines,
                                                                part_elset_data,
                                                                material_data)

    text_info = text_info + "\n\n{:>10s}".format("cell #")
    text_info = text_info + " {:>10s} {:>20s} {:>50s}".format("mat #", "density", "mat_name")
    text_info = text_info + " {:>50s} {:>50s} {:>50s}".format('elset_name', 'part_name', 'instance_name')
    text_info = text_info + "\n"
    if dentype == 'mass_den':
        text_info = text_info + " {:>20s} {:>20s}".format("", "(g/cm^3)")
    elif dentype == "atom_den":
         text_info = text_info + " {:>20s} {:>20s}".format("", "(atoms/barn-cm)")
    else:
        raise ValueError('>>>Error: invalid density unit')
   
    if len(heading) > 0:
        if len(heading) <= 128:
            text = heading
        else:
            text = heading[0:127]
    else:
        text = "No description"
    text = text + "\nC"
    text = text + "\nC Abaqus Input File: {}".format(filein)
    text = text + "\nC"
    text = text + "\nC High-Level Mesh Information:"
    text = text + "\nC   Number of Parts:     {:d}".format(len(part_names))
    text = text + "\nC   Number of Instances: {:d}".format(len(instance_names))
    text = text + "\nC   Number of Materials: {:d}".format(len(material_names))
    text = text + "\nC"
    text = text + "\nC Details of mesh information are in file: {:s}".format(meshinfo)
    text = text + "\nC"
    text = text + "\nC PSEUDO CELLS"

    sp = '-'
    cell_num   = 0
    cell_mat_names = []
    
    for ins_name in instance_names:
        ins_data   = instance_data.get(ins_name)
        pname      = ins_data[0]
        
        pelsetdata       = part_elset_data.get(pname)
        order_elset_data = pelsetdata[0]
        check_key        = pelsetdata[2]

        for elset in order_elset_data:
            key_words = elset[0]
            if check_key not in key_words: continue

            elset_num   = elset[1]
            name        = '{:s}-{:d}'.format(pname,elset_num)

            pelsetmat   = part_elset_material_data.get(name)
            mat_name    = pelsetmat[0]
            mat_num     = pelsetmat[1]
            mat_density = pelsetmat[2]

            elset_name = sp.join(key_words) + '-{:d}'.format(elset_num) 
            cell_num   = cell_num + 1
               
            text_info = text_info + "\n{:10d}".format(cell_num)
            if dentype == 'mass_den':
                text_info = text_info + " {:>10d} {:>20.6f} {:>50s}".format(mat_num, mat_density, mat_name)
            elif dentype == 'atom_den':
                text_info = text_info + " {:>10d} {:>20.6e} {:>50s}".format(mat_num, mat_density, mat_name)
            else:
                raise ValueError('>>>Error: invalid density unit')
            text_info = text_info + " {:>50s} {:>50s} {:>50s}".format(elset_name, pname, ins_name)

            if writecomments:
                text = text + '\nC\nC {:d}: {:s}, {:s}, {:s}'.format(cell_num, ins_name, pname, elset_name)
                
            if dentype == 'mass_den':
                text = text +'\n{:d} {:>8d} {:>12.6f}      0  u=1  $ {:s}'.format(cell_num, mat_num, -mat_density,mat_name)

            elif dentype == 'atom_den':
                text = text +'\n{:d} {:>8d} {:>12.6e}      0  u=1  $ {:s}'.format(cell_num, mat_num, mat_density,mat_name)
            
            else:
                raise ValueError('>>>Error: invalid density unit')

            cell_mat_names.append(mat_name)
                
    if bgmaterial != 0:
        bgmaterial_data = material_data.get(bgmaterial)
        if bgmaterial_data is None:
            print(">>>Warning: background material {:d} is not in the list of materials".format(bgmaterial))
            print("            set background material to 0")
            bgmaterial = 0
            bgden      = None
        else:
            bgden = bgmaterial_data[1]

    else:
        bgden = None

    if writecomments: text = text+'\nC'
        
    if bgden is None:
        text = text + '\n{:d} {:>8d} {:>12s}      0  u=1  $ background'.format(cell_num+1,bgmaterial,'')
    else:
        if dentype == 'mass_den':
            bgden = -bgden
            text = text +'\n{:d} {:>8d} {:>12.6f}      0  u=1  $ background'.format(cell_num, bgmaterial, bgden)
        elif dentype == 'atom_den':
            text = text +'\n{:d} {:>8d} {:>12.6e}      0  u=1  $ background'.format(cell_num, bgmaterial, bgden)

    mat_num = 0
    text = text + '\nC'
    text = text + '\nC LEGACY CELLS'
    text = text + '\n{:d} {:>8d} {:>12s}     -999  fill=1 $ fill cell'.format(cell_num+2, mat_num, '')

    if include_cell_cards:
        f = open(cellcards,'r')
        lines = f.readlines()
        f.close()
        text  = text + "\nC\n"
        text = text + ''.join(lines)
    else:
         text = text + '\n{:d} {:>8d} {:>12s}      999'.format(cell_num+3,mat_num,'')
            
    text = text + '\n'
    text = text + '\nC Surface Cards'
    text = text + '\nC'
    text = text + '\n999 sph {:15.6E} {:15.6E} {:15.6E} {:15.6E}'.format(center[0], center[1], center[2], radius)
    if include_surface_cards:
        f = open(surfacecards,'r')
        lines = f.readlines()
        f.close()
        text  = text + "\nC\n"
        text  = text +''.join(lines)
         
    text = text+'\n'
    text = text+'\nC Data Cards'

    mat_name = cell_mat_names[0]
    text2= "1 1  $ {:s}".format(mat_name)
    text3= ''
    for i in range(1,cell_num):
        mat_name = cell_mat_names[i]
        text3 = text3 +  "\n               {:d} {:d}  $ {:s}".format(i+1,i+1,mat_name)
        
    text = text + "\nC"
    text = text + "\nembed1 meshgeo=abaqus"
    text = text + "\n       mgeoin={:s}".format(filein)
    text = text + "\n       meeout={:s}".format(eeout)
    text = text + "\n       length={:E}".format(lenconv)
    text = text + "\n       background={:d}".format(cell_num+1)
    text = text + "\n       matcell={:s}".format(text2)
    text = text + text3
    
    if include_data_cards:
        text = text +'\nC\n'
        f = open(datacards,'r')
        lines = f.readlines()
        f.close()
        text = text +''.join(lines)
    else:
        text = text+'\n'
            
    # write MCNP skeleton input file
    f = open(fileout,'w')
    f.write(text)
    f.close()
    print("\n   MCNP input file: {:s}".format(fileout))
    
    # write unstructured mesh information file
    if writemeshinfo:
        text_info = text_info + '\n\n' + mesh_info
    f = open(meshinfo,'w')
    f.write(text_info)
    f.close()
    if writemeshinfo:
        print("   Mesh information file: {:s}".format(meshinfo))

    print('\n')

    return text, text_info

#------------------------------------------------------------------------------
if __name__=='__main__':
    import argparse

    parser = argparse.ArgumentParser(description='** Write MCNP Unstructured Mesh Input file **')

    parser.add_argument("-i", "--input", required=True, metavar="<file.inp>", 
                        type=str,
                        help="Abaqus input for constructing MCNP input file")
    parser.add_argument("-o", "--output", metavar="<file.mcnp>",
                        type=str,
                        help="output file name")
    
    parser.add_argument("-e", "--eeout",metavar="<file.eeout>",
                        type=str,
                        help="eeout file name in EMBED card")
    parser.add_argument('-mi','--meshinfo', metavar='<filename.info>',
                        type=str,
                        help='a mesh information file name')
    
    parser.add_argument("-cc", "--cellcards", metavar="<cellcards.txt>",
                         type=str,
                         help='MCNP cell cards file to include')
    
    parser.add_argument("-sc", "--surfacecards", metavar="<surfacecards.txt>",
                         type=str,
                         help='MCNP surface cards file to include')
    
    parser.add_argument("-dc", "--datacards", metavar="<datacards.txt>",
                         type=str,
                         help='MCNP data cards file to include')

    parser.add_argument("-b", "--back", metavar="<material_number>",
                        default=0, type=int,
                        help='background material for MCNP input file')
    
    parser.add_argument("-l", "--length", metavar="<len_conversion>", 
                        default=1.0, type=float,
                        help="a multiplication conversion factor to centimeters")
    
    parser.add_argument('-re', '--radext', metavar="<radius_extension>",
                        default=1.0, type=float,
                        help="a radius extension of a sphere surface for a fill cell (in centimeters)")
    
    parser.add_argument("-du", "--densityunit", metavar="<density_unit>",
                         default='g/cm^3', type=str,
                         help='density unit option: g/cm^3 [default] or atoms/barn-cm')
    
    parser.add_argument("-wc", "--writecomments",
                        action='store_true',
                        help='write pseudo-cell comments in MCNP input file')
    
    parser.add_argument("-wm", "--writemeshinfo",
                        action='store_true',
                        help='write mesh information into mesh information file')
    

    args = parser.parse_args()

    if args.input is not None:
        write_mcnp_um_input(filein=args.input,
                            fileout=args.output, 
                            eeout=args.eeout, 
                            meshinfo=args.meshinfo,
                            cellcards=args.cellcards,
                            surfacecards=args.surfacecards,
                            datacards=args.datacards,
                            bgmaterial=args.back,
                            lenconv=args.length,
                            radiusext=args.radext,
                            denunit=args.densityunit,
                            writemeshinfo=args.writemeshinfo,
                            writecomments=args.writecomments)
        
                   
            
             
    
                 
                   
            
    


                        
                  

                    
                    
                

                
                
                


